#include "GenomeMem.h"

/*
 * Constructor.  Creates new Genome with filename.
 * fn - the filename of the file to be parsed into the genome.
 */
Genome::Genome(const char* fn) :
	offset(0), bin_offset(0), genome(NULL) {
	Init(fn);
	reader = new Reader();
}

Genome::Genome() :
	offset(0), bin_offset(0), reader(NULL), genome(NULL) {

}

Genome::~Genome() {
	delete reader;
} 

void Genome::Init(const char* fn) {
	// Don't need to init names if we're going to do so in the read file
	if(!gREAD) 	{
		if(names.size()) //if we've already set up the "names" vector, just return.
			return;

		char files[strlen(fn)];
		strcpy(files,fn);
		
		const char* delims = ", \"\n";
		char* token = strtok(files,delims);
		if(VERBOSE > 1)
			cout << "Found:" << endl;

		while(token != NULL) {
			if(VERBOSE)
				cout << "\t" << token << endl;
			string temp_str = token;
			names.push_back(pair<string,unsigned long>(temp_str,0));
			
			token = strtok(NULL,delims);
		}

		//Push a dummy string on the end so we can know when to end...
		names.push_back(pair<string,unsigned long>("end",0));
	}

	//clear the *_spots files
	for(int i=0; i<NUM_CHAR_VALS; i++) {
		reads[i] = 0;
		gens[i] = 0;
	}

}

void Genome::use(const char* fn) {
	if(reader != NULL) {
		cout << "non-null reader" << endl;
		delete reader;
	}
	Init(fn);
	reader = new Reader();
}

void Genome::print_to_file(hash_map<unsigned int, HashLocation> & gh, char* ofn) { 
	printf("Printing to file: %s\n",ofn);
	int SEQ_LENGTH = 35; //This is an archaic #def'd element that needed to be removed.
	ofstream out;
	out.open(ofn,ios_base::out);
	
	bin_seq bs;
	
	unsigned int i;
	
	for(i=0; i<gHASH_SIZE; ++i) {
		if(!gh[i].size)
			continue;
		out << i << "\t" << gh[i].size;
		unsigned int j;
		out << '\t' << bs.hash2str(i,MER_SIZE);
		for(j=0; j<gh[i].size; j++)
			out << '\t' << gh[i].hash_arr[j] << "[" << GetPos(gh[i].hash_arr[j],true) << "] :" 
				<< GetString(gh[i].hash_arr[j],SEQ_LENGTH);
		out << endl;
	}
	
	out.close();
}

unsigned int Genome::readGen(char* fn) {
	FILE* readF;

	readF = fopen(fn,"r");
	
	if(!readF) {
		printf("Trying to open: >%s<\n",fn);
		perror("File wouldn't open");
		throw "Could not open Genome read file";
	}

	unsigned long temp_gen_hash_size;
	//fscanf(readF,"%ld\n",&gen_hash_size);
	fread(&temp_gen_hash_size, sizeof(temp_gen_hash_size), 1, readF);
	gHASH_SIZE = temp_gen_hash_size;
	// fprintf(stderr,"gen_hash_size: %d\n",gHASH_SIZE);
	gen_hash.set_size(gHASH_SIZE);

	// Read in the HASH_SIZE
	fread(&MER_SIZE, sizeof(MER_SIZE), 1, readF);
	// MER_SIZE = 10;
	fprintf(stderr,"\tMer Size: %d\n",MER_SIZE);
	
	// Read in the hash
	while(true) {
		unsigned int index,hash_size;
		//fscanf(readF,"%d %d\n",&index,&hash_size);
		fread(&index, sizeof(index), 1, readF);
		fread(&hash_size, sizeof(hash_size), 1, readF);
		//fprintf(stderr,"%d %d\n",index,hash_size);
		if(!hash_size)	//This is a marker for the end of the read.
			break;

		gen_hash[index].size = hash_size;
		// malloc a new array for this location
		unsigned long* temp_arr = new unsigned long[hash_size];

		for(unsigned int j=0; j<hash_size; j++) {
			unsigned long hash_spot;
			fread(&hash_spot,sizeof(unsigned long),1,readF);
			temp_arr[j] = hash_spot;
			//gen_hash[index].push_back(hash_spot);
		}
		gen_hash[index].hash_arr = temp_arr;
		if(hash_size != gen_hash[index].size)
			fprintf(stderr,"****At location %d, %ld should be %d\n",index,gen_hash[index].size,hash_size);
	}

	// fprintf(stderr,"Finished reading in hash.  Reading in the genome now\n");
		
	// Read in the genome.
	unsigned long temp_genome_size = 0;

	GEN_TYPE former_gen_size = GEN_SIZE;
	// The GEN_SIZE is the number of chars stored in each GenomeLocation
	fread(&GEN_SIZE, sizeof(GEN_SIZE), 1, readF);
	// The size of the genome
	fread(&temp_genome_size, sizeof(temp_genome_size), 1, readF);
	if(former_gen_size != GEN_SIZE)
		fprintf(stderr,"\tGEN_SIZE: %d\n",GEN_SIZE);

	genome = new GenomeLocation[temp_genome_size];

	for(unsigned int i=0; i<temp_genome_size; i++) {
		GEN_TYPE genome_char = 0;
		fread(&(genome_char),sizeof(GEN_TYPE),1,readF);
		genome[i].packed_char = genome_char;
		genome[i].amount = 0;
		// genome.push_back(pair<GEN_TYPE,float>(genome_char,0));
	}

	genome_size = temp_genome_size;

	int num_chrs = 0;
	fscanf(readF,"\n%d\n",&num_chrs);
	//fprintf(stderr,"\n%d\n",num_chrs);
	// Write the chromosome names to a file
	for(int i=0; i<num_chrs; i++) {
		char name[255];
		unsigned long chr_pos;
		fscanf(readF,"%s %ld\n",name,&chr_pos);
		//fprintf(stderr,"%s %ld\n",name,chr_pos);
		names.push_back(pair<string,unsigned long>(name,chr_pos));
	}
	
	fclose(readF);

	return 0;
}


void Genome::LoadGenome() {
	if(gREAD) {
		fprintf(stderr,"\tReading from file: %s\n",gREAD_FN);// read the file in from memory
		readGen(gREAD_FN);
		fprintf(stderr,"\tJust finished reading file %s.\n",gREAD_FN);
	}
	else
		hash_and_store();


	//print_to_file(gen_hash,"printed_hash");
	fix_hash();


	//if we are doing the bisulfite sequencing, we want to record what nucleotide
	//comes at each position
	if(BISULFITE) {
		printf("Using bisulfite conversion techniques\n");
		for(unsigned int i=0; i<NUM_CHAR_VALS; i++)
			reads[i] = new float[genome_size];
		
		gens[1] = new float[genome_size]; //Represents a genomic 'c'
		//gens[3] = new float[genome_vec.size()];
	}
	if(ATOG) {
		for(unsigned int i=0; i<NUM_CHAR_VALS; i++)
			reads[i] = new float[genome_size];
		
		gens[0] = new float[genome_size];	// Represents a genomic 'a'
		//gens[2] = new float[genome_vec.size()];
	}

	if(gSAVE) {
		fprintf(stderr,"\tSaving to file: %s\n",gSAVE_FN);
		saveGen(gSAVE_FN);
	}
}

void Genome::hash_and_store() {
	
	// Set some values that are necessary for the 'store()' function
	offset = 0;
	bin_offset = 0;

	hash_map<unsigned int, vector<unsigned long> > gen_hash_vec;
	Reader reader_vec;
	vector<pair<GEN_TYPE,float> > genome_vec;
	unsigned char gen_piece[BUFFER_SIZE+1];

	printf("Hashing Genome...\n");
	gen_hash_vec.set_size(gHASH_SIZE);

	
	bin_seq bs;	

	unsigned long counter = 0;
	unsigned long int chars_stored = 0;
	unsigned long int j_pos = chars_stored-0;
	unsigned int j;

	for(j=0; j<names.size()-1; j++) {
		reader_vec.use(names[j].first);
		string chr_name = reader_vec.GetName();
		//unsigned long j_pos = counter;
		j_pos = chars_stored;
						
		unsigned long file_size = reader_vec.GetFileSize();
		//cout << "File Size: " << file_size << endl;
		bool keep_reading = true;
		unsigned int i;

		while(keep_reading) { 
	
			keep_reading = reader_vec.read(gen_piece);
	
			//If we're not at the end of the file, then we can throw it the whole chunk,
			//without the extra MER_SIZE bases at the end.
			//When we come to the end of the file, we can give it everything.
			//We'll also return the number of characters stored.
			if(keep_reading) {
				chars_stored += store(genome_vec,gen_piece,BUFFER_SIZE-MER_SIZE);
				//printf("keeping reading\n");
			}
			else
				chars_stored += store(genome_vec,gen_piece,BUFFER_SIZE);
	 
			for(i=0; i < BUFFER_SIZE-MER_SIZE; i++) {
					
				pair<int ,unsigned int> hash = bs.get_hash(i,gen_piece);

				if(hash.first >= 0) {	//if it's zero, that's a result of an N character.
					if(hash.first != 0)	{	//don't want to store N's.
						if(counter % GEN_SKIP == 0) //This allows us a smaller hash size in total.
							gen_hash_vec[hash.second].push_back(counter);
						
						// print output
						if( VERBOSE && ((counter % (file_size/50) ) == 0)) {
							cout << ".";
							cout.flush();
						}

					}
				}
						
				//if hash.first is -2, it means it's a new genome.
				if(hash.first == -2) {
					
					//Find the name of this genome.
					//at position i, we should have the '>'
					char gen_name[100];	//the name can only be 100 chars long
					unsigned int gen_pos=0; 
					unsigned int name_pos = i;
					
					//we want to read the name of the next chromosome that's been found
					while(gen_piece[++i] != '\n' && i < BUFFER_SIZE) { 
						if(gen_pos < 100)
							gen_name[gen_pos] = gen_piece[i];
						gen_pos++;
					}
					//Three things could happen here: 
					// 1. The name occurs at the end of the buffer
					// 2. The name occurs accross buffers
					// 3. The name occurs near the end of the buffer so you don't read
					//	  the last few bases of the sequence.
					
					//*** SPECIAL CASE--only happens very rarely ***//
					if(i >= BUFFER_SIZE) {	//it didn't finish reading the name
						char* line = reader_vec.ReadLine();
						string temp = line;
						delete[] line;
						
						// also need to store any potentially missed characters
						// We want to include the name so we can find it...
						// But if the name portion is short, we want to at least the last
						// MER_SIZE bases
						if(BUFFER_SIZE-name_pos > (unsigned int)MER_SIZE) 
							store(genome_vec,gen_piece+name_pos,BUFFER_SIZE-name_pos);
						else
							store(genome_vec,gen_piece+BUFFER_SIZE-MER_SIZE,MER_SIZE);
						
						int chr_name_pos = 0;
						while(gen_pos < 100)	//don't overlow the gen_name
							gen_name[gen_pos++] = temp[chr_name_pos++];
					}
					
					//*** SPECIAL CASE #3 ***//
					// Here, there's not enough characters on the end of the sequence to read
					// a new hash, so it throws off the numbering and reading(?).
					else if(i >= BUFFER_SIZE - MER_SIZE) {
						// Shift the reader's offset by the number of bases remaining in the sequence
						// so it will read them again.
						reader_vec.ShiftOffset(BUFFER_SIZE - i);
					}

					gen_name[gen_pos] = '\0'; 
					string str_gen_name = gen_name;
					//printf("Found new genome at pos %u: %s\n",i,gen_name);
					if(VERBOSE > 1) 
						printf("Found new genome at pos %u: %s\n",i,gen_name);

					vector<pair<string,unsigned long> >::iterator vit = names.begin();
					//Find the position of this genome name
					for(unsigned int name_pos=0; name_pos < j; name_pos++)
						vit++;
					vit++;	// the vector inserts to the one "previous", but we want to insert
							// after the current name.

					//printf("Inserting: %s, %d at %d before %s, %d ",str_gen_name.c_str(),0,j,(*vit).first.c_str(),(*vit).second);
					//printf("and after %s, %d",(*(vit-1)).first.c_str(),(*(vit-1)).second);
					names.insert(vit,pair<string,unsigned long>(str_gen_name,0));

					//take care of the former name.
					//1. set the name in the genome to have the correct position
					//2. set the j_pos var to be the number of chars we've stored so far.
					names[j] = pair<string,unsigned long>(chr_name,j_pos);
					chr_name = str_gen_name;
					
					counter += hash.second;	//increment the counter the number of bases we've skipped

					j_pos = counter;
					j++;
				}
				

				//if hash.first is -1, it means it's found the end-of-line char.
				else if(hash.first == -1 || i >= BUFFER_SIZE-MER_SIZE) {
					//if(hash.first == -1)
					//	printf("Found the eof char\n");
					break;
				}
				else	// can't be -2 or -1, so it must be nonnegative.
					counter++;
			}
		}
		if(VERBOSE)
			cout << endl;

		names[j] = pair<string,unsigned long>(chr_name,j_pos);
		
		//there are cases where the end of the line throws off our numbering
		//We'll fix it here.
		unsigned long int end = i;
		int count=0;
		for(unsigned int p=end; gen_piece[p] != '\0'; p++) {
			//if(base2int(gen_piece[p]) <= 4) {
			if(gen_CONVERSION[(int)gen_piece[p]] <= 4) {
				counter++;
				count++;
			}
		}
	}
	j_pos = chars_stored;
	names[j] = pair<string,unsigned long>("end",j_pos);
	
	// cerr << "Size of genome: " << genome_vec.size() * GEN_SIZE << endl;

		
#ifdef DEBUG
		for(unsigned int i=0; i<names.size(); i++) {
			printf("Name: %s\tBegin: %ld\n",names[i].first.c_str(),names[i].second);
		}
#endif

	// Need to convert it to the real, vector genome.
	convertToVector(gen_hash_vec, genome_vec);


	// fprintf(stderr,"End of hash_and_store\n");
}

void Genome::convertToVector(hash_map<unsigned int, vector<unsigned long> > &gh, 
							 vector<pair<GEN_TYPE, float> > &g) {
	
	// Set up the hashed genome
	gen_hash.set_size(gh.size());
	for(unsigned int i=0; i<gh.size(); i++) {
		gen_hash[i].size = gh[i].size();
		unsigned long* temp_arr = new unsigned long[gh[i].size()];
		for(unsigned int j=0; j<gh[i].size(); j++) {
			temp_arr[j] = gh[i][j];
		}
		gen_hash[i].hash_arr = temp_arr;

		// Save space by clearing this location.
		gh[i].clear();
	}

	// Set up the genome
	if(genome)
		delete[] genome;
	genome = new GenomeLocation[g.size()];
	for(unsigned int i=0; i<g.size(); i++) {
		genome[i].packed_char = g[i].first;
		genome[i].amount = 0;
	}
	genome_size = g.size();
}

/*
 * store() will store the genome into a smaller array of BIT64_t portions.
 * Each character is represented by 4 bits (because we require the 4 nucleotides
 * and an additional N character), and can be accessed directly.  
 *
 * One of the problems we have now is the last few characters in gen.  They will
 * appear the second time at the beginning of the genome yet again, throwing off the 
 * numbering system. 
 *
 * @par gen - the piece of the genome to store.
 * @par num - the number of characters to store.
 */
unsigned long int Genome::store(vector<pair<GEN_TYPE,float> > &g, unsigned char* gen, 
								unsigned long int num) {
	unsigned long gen_pos=0;
	unsigned long int num_bases=0;
	
	while(gen_pos<num) {
		GEN_TYPE temp = 0;
		
		if(bin_offset != 0) {
			temp = g[offset].first;
			g.pop_back();
		}
			
		for(; (bin_offset<GEN_SIZE) && gen_pos<num; gen_pos++) {	//can fit 16 chars into a BIT64_t 
			//int int_to_add = base2int(gen[gen_pos]);
			int int_to_add = gen_CONVERSION[(int)gen[gen_pos]];
			
			
			//If it's a valid char.
			if(int_to_add < 5) {
				num_bases++;
				temp <<= 4;
				temp |= int_to_add;
				++bin_offset;
			}
			//If we come accross the flag for the end of the stream--should never happen...
			else if(int_to_add == 6) { 
				gen_pos = num;
				//cout << "Found end." << endl;
				break;
			}
			else if(int_to_add == 7) {	//come accross a new chromosome name.
				string temp = (char*)(gen+gen_pos);
				while(gen[gen_pos] != '\n' && gen[gen_pos] != '\0')
					gen_pos++;	//increment the counter until we've found the end of the name
			}
			else ;//do nothing--it's an invalid character and not an end-of-line
		}
		if(bin_offset != 0)	{ //if we've made any changes to this value...
			g.push_back(pair<GEN_TYPE,float>(temp,0));
		}
		else ;	//no changes.  Coming out of the previous for-loop
		
		
		//means we've filled it fully--so increment the offset.
		if(bin_offset == GEN_SIZE) {
			offset++;
			bin_offset = 0;
		}


	}
	
	//printf("Just read %ld bases\n",num_bases);
	return num_bases;
}

/*
 * This particular store function is used for the DataGen program--not for genhash. 
 * It is meant to store the genome without needing to hash every member--speeds up the proccess
 * a bit.
 */
void Genome::store() {
	
	hash_map<unsigned int, vector<unsigned long> > gen_hash_vec;
	Reader reader_vec;
	vector<pair<GEN_TYPE,float> > genome_vec;
	unsigned char gen_piece[BUFFER_SIZE+1];

	if(VERBOSE)
		cerr << "Storing the genome." << endl;

	gen_hash_vec.set_size(gHASH_SIZE);

	
	// We dont need to use a set--the members are already sorted.  And a vector would have
	// faster random access than a list, which is a plus.
	// hash_map<unsigned int, vector<unsigned int> > gen_hash;
	
	bin_seq bs;	

	unsigned long counter = 0;
	unsigned long int chars_stored = 0, j_pos = 0;
	unsigned int j;

	for(j=0; j<names.size()-1; j++) {
		reader_vec.use(names[j].first);
		string chr_name = reader_vec.GetName();
		j_pos = chars_stored;
					
		unsigned long file_size = reader_vec.GetFileSize();
		unsigned int i;
		bool keep_reading = true;
		
		while(keep_reading) { 
	
			keep_reading = reader_vec.read(gen_piece);
	
			//If we're not at the end of the file, then we can throw it the whole chunk,
			//without the extra MER_SIZE bases at the end.
			//When we come to the end of the file, we can give it everything.
			if(keep_reading)
				chars_stored += store(genome_vec, gen_piece, BUFFER_SIZE-MER_SIZE);
			else
				chars_stored += store(genome_vec, gen_piece, BUFFER_SIZE);
	
			for(i=0; i < BUFFER_SIZE-MER_SIZE; i++) {
									
				pair<int,unsigned int> hash = bs.get_hash(i,gen_piece);

				if(hash.first >= 0) {
					if(hash.first)
						//if(counter % 3 == 0) //This allows us a smaller hash size in total.
							gen_hash_vec[hash.second].push_back(counter);
					counter++;
				}
						
				if( VERBOSE && hash.first && ((counter % (file_size/50) ) == 0)) {
					cout << ".";
					cout.flush();
				}
				
				
				//if hash.first is -2, it means it's a new genome.
				if(hash.first == -2) {
					
					//Find the name of this genome.
					//at position i, we should have the '>'
					char gen_name[100];	//the name can only be 100 chars long
					int gen_pos=0; 
					while(gen_piece[++i] != '\n' && i < BUFFER_SIZE) { 
						gen_name[gen_pos] = gen_piece[i];
						gen_pos++;
					}
					gen_name[gen_pos] = '\0'; 
					string str_gen_name = gen_name;
					//printf("Found new genome at pos %u: %s\n",i,gen_name);

					vector<pair<string,unsigned long> >::iterator vit = names.begin();
					//Find the position of this genome name
					for(unsigned int name_pos=0; name_pos < j; name_pos++)
						vit++;
					vit++;	//the vector inserts to the one "previous", but we want to insert
							//after the current name.

					names.insert(vit,pair<string,unsigned long>(str_gen_name,0));

					//take care of the former name.
					//1. set the name in the genome to have the correct position
					//2. set the j_pos var to be the number of chars we've stored so far.
					names[j] = pair<string,unsigned long>(chr_name,j_pos);
					chr_name = str_gen_name;
					
					counter += hash.second;	//increment the counter the number of bases we've skipped
					counter -= 1;

					j_pos = counter;
					j++;
				}
				

				//if hash.first is -1, it means it's found the end-of-line char.
				if(hash.first == -1 || i >= BUFFER_SIZE-MER_SIZE) {
					break;
				}

			}
		}
		if(VERBOSE)
			cout << endl;
		names[j] = pair<string,unsigned long>(chr_name,j_pos);

		//there are cases where the end of the line throws off our numbering.
		//We'll fix it here.

		unsigned int end = i;
		int count=0;
		for(unsigned int p=end; gen_piece[p] != '\0'; p++) {
			//if(base2int(gen_piece[p]) <= 4) {
			if(gen_CONVERSION[(int)gen_piece[p]] <= 4) {
				counter++;
				count++;
			}
		}

	}

	j_pos = chars_stored;
	names[j] = pair<string,unsigned long>("end",j_pos);

#ifdef DEBUG
	for(unsigned int i=0; i<names.size(); i++) {
		printf("Name: %s\tBegin: %ld\n",names[i].first.c_str(),names[i].second);
	}
#endif

	// Convert this to a vector thing
	convertToVector(gen_hash_vec, genome_vec);
}


unsigned int Genome::saveGen(char* fn) {
	FILE* saveF;

	saveF = fopen(fn,"w");
	if(!saveF)
		throw "Could not open Genome save file";

	// Write the hash out to a file
	//fprintf(saveF,"%ld\n",gen_hash.size());
	unsigned long tempsize = gen_hash.size();
	fwrite(&tempsize, sizeof(unsigned long), 1, saveF);
	fwrite(&MER_SIZE, sizeof(int), 1, saveF);
	//fprintf(stderr,"%ld\n",gen_hash.size());
	for(unsigned int i=0; i<gen_hash.size(); i++) {
		if(gen_hash[i].size) {
			unsigned int hash_size = gen_hash[i].size;
			//fprintf(saveF,"%d %d\n",i,hash_size);
			fwrite(&i, sizeof(unsigned int), 1, saveF);
			fwrite(&hash_size, sizeof(unsigned int), 1, saveF);
			//fprintf(stderr,"%d %d\n",i,hash_size);
			for(unsigned int j=0; j<hash_size; j++) {
				fwrite(&(gen_hash[i].hash_arr[j]),sizeof(unsigned long),1,saveF);
			}
		}
	}
	// Write out an end marker
	//fprintf(saveF,"0 0\n");
	unsigned int tempmarker = 0;
	fwrite(&tempmarker, sizeof(unsigned int), 1, saveF);
	fwrite(&tempmarker, sizeof(unsigned int), 1, saveF);

	// Write the genome out to a file.
	//fprintf(saveF,"%d %ld\n",GEN_SIZE,genome.size());
	unsigned long tempgenome_size = genome_size;
	fwrite(&GEN_SIZE, sizeof(GEN_SIZE), 1, saveF);
	fwrite(&tempgenome_size, sizeof(tempgenome_size), 1, saveF);
	//fprintf(stderr,"%d %ld\n",GEN_SIZE,genome.size());
	for(unsigned int i=0; i<genome_size; i++) {
		fwrite(&(genome[i].packed_char),sizeof(GEN_TYPE),1,saveF);
	}

	fprintf(saveF,"\n%d\n",(int)names.size());
	//fprintf(stderr,"\n%d\n",(int)names.size());
	// Write the chromosome names to a file
	for(unsigned int i=0; i<names.size(); i++) {
		fprintf(saveF,"%s %ld\n",names[i].first.c_str(),names[i].second);
	}
	
	fclose(saveF);

	return 0;
}

void Genome::fix_hash() {
	//Means there is no user-defined max-hash.
	if(MAX_HASH_SIZE <= 0)
		return;

	for(unsigned int i=0; i<gHASH_SIZE; i++)
		if(gen_hash[i].size > MAX_HASH_SIZE) {
#ifdef DEBUG
			if(VERBOSE > 1)
				printf("Clearing hash at pos %d with size %d\n",i,gen_hash[i].size);
#endif
			delete[] gen_hash[i].hash_arr;
		}
}


/*
 * Returns a string from begin to begin+size.
 * the first position in the genome is position 0.
 * @par begin - the beginning of the string to return
 * @par size - the size of the string to return.
 */
string Genome::GetString(const unsigned long begin, const unsigned int size) { 
	string to_return = "";

	/***********************Check for Bounds******************************/
	//We don't want to get a string beyond the end of the chromosome...
	int chr_num=0;
	//											it's already been pre-incremented
	while(names[chr_num].second <= begin && ((unsigned int)chr_num <= names.size()) ) chr_num++;
	//while(names[++chr_num].second <= begin && ((unsigned int)chr_num <= names.size()) );
	
	//means it's not valid--will extend beyond the chromosome file
	if( (names[chr_num].second < begin+size) ) 
		//printf("Extending Beyong the Chromosome\n");
		return "";
	/*********************************************************************/

	unsigned long pos_in_gen = begin/GEN_SIZE; 
	unsigned int temp_size = size;
	
	GEN_TYPE t_temp = genome[pos_in_gen].packed_char;

	//cout << "\tbegin: " << begin << endl;
	//cout << "\tpos_in_gen: " << pos_in_gen << endl;
	//cout << "\tt_temp: " << t_temp << endl;
		
	//if(VERBOSE > 3)
	//	cout << "getting position: " << t_temp << endl;
	
	unsigned int count=0;

	while(count < size) {

		string temp = "";
	
		if(count == 0) {	//get the last few characters off the BIT64_t
			unsigned long bit_offset = (begin%GEN_SIZE);

			//for(unsigned long j=0; j<GEN_SIZE-bit_offset; j++) {
			for(unsigned int j=GEN_SIZE-bit_offset; j>0; j--) {
				//shift 0xf left 4 bases for each j.
				//then we don't need to reverse the string.
				//printf("Shifting by: %d = %lld\n",4*(j-1),(t_temp>>(4*(j-1))) & 0xf);
				//temp += int2base(t_temp & 0xf);
				//temp += int2base(0xf & (t_temp >> (4*(j-1))));
				temp += INT2BASE[0xf & (t_temp >> (4*(j-1)))];
				//t_temp >>= 4;
				count++;
			}
			//temp = reverse(temp);
			if(temp.size() > size)
				return temp.substr(0,size);
		}
		else {	//we now want the first offset characters off the BIT64_t.
			//for(unsigned int j=0; j<GEN_SIZE; j++) {
			for(unsigned int j=GEN_SIZE; j>0; j--) {
				
				//printf("Shifting by: %d = %lld\n",4*(j-1),(t_temp>>(4*(j-1))) & 0xf);
				//temp += int2base(t_temp & 0xf);
				//temp += int2base((t_temp >> (4*(j-1))) & 0xf);
				temp += INT2BASE[(t_temp >> (4*(j-1))) & 0xf];
				//t_temp >>= 4;
			}
			//temp = reverse(temp);
			temp = temp.substr(0,temp_size);
			count += temp.size();
		}
		to_return += temp;
		
		temp_size = size - count;
		
		t_temp = genome[++pos_in_gen].packed_char;
	}

	return to_return;
}

inline int Genome::base2int(char c) {
	switch(c) {
		case 'A' : return 0;
		case 'C' : return 1;
		case 'G' : return 2;
		case 'T' : return 3;
		case 'a' : return 0;
		case 'c' : return 1;
		case 'g' : return 2;
		case 't' : return 3; 
		case 'N' : return 4;
		case 'n' : return 4;
		case '\n' : return 5;
		case '\r' : return 5;
		case '\0' : return 6;	//6 is the flag for the end of the genome stream.
		case '>' : return 7;
		default: return 4;	//we want invalid characters to be represented by "N"
	}
	return 5;	//5 is an invalid character.
} 

inline char Genome::int2base(int i) {
	//printf("i: %d\n",i);
	switch(i) {
		case 0: return 'a';
		case 1: return 'c';
		case 2: return 'g';
		case 3: return 't';
		case 4: return 'n';
	}
	return '?';
}

inline string Genome::reverse(string &s) {
	string temp = "";
	int len = s.length();
	
	for(int i=1; i<=len; i++)
		temp += s[len-i];
	
	return temp;
}

HashLocation* Genome::GetMatches(unsigned int hash) {
	return &gen_hash[hash];
}

float Genome::GetScore(const unsigned long &pos) {
	unsigned long pos_temp = pos/GEN_SIZE;
	return genome[pos_temp].amount;
}
void Genome::AddScore(const unsigned long &pos, const float &amt) {
	//float pos_temp = pos / BIN_SIZE;
	//pos_temp *= BIN_SIZE;
	unsigned long pos_temp = pos;
	genome[pos_temp/GEN_SIZE].amount+= amt;
}

void Genome::AddSeqScore(const unsigned long pos, const float amt, char which) {
	switch(which) {
		case 'a':
			if(reads[0])	//if we've created one, then add to it.  Used for both BISULFITE and ATOG
				reads[0][pos/GEN_SIZE] += amt;
			break;
		case 'c':
			if(reads[1])
				reads[1][pos/GEN_SIZE] += amt;
			break;
		case 'g':
			if(reads[2])
				reads[2][pos/GEN_SIZE] += amt;
			break;
		case 't':
			if(reads[3])
				reads[3][pos/GEN_SIZE] += amt;
			break;
		case '-':
			if(reads[4])
				reads[4][pos/GEN_SIZE] += amt;
			break;
	}
}

void Genome::AddGenScore(const unsigned long pos, const float amt, char which) {
	//genome[pos/GEN_SIZE].second += amt;	//we only want to add it to the genome once.
										//we'll do it while we're adding to the genome sequence files.
							//actually, we're already adding the score of each sequence to the genome.
							//we're doing it a second time here.  bad news, bears.
	switch(which) {
		case 'a':
			if(gens[0])
				//gens[0][pos/GEN_SIZE] += amt;
				gens[0][pos/GEN_SIZE] = 1;
		break;
		case 'c':
			if(gens[1])
				//gens[1][pos/GEN_SIZE] += amt;
				gens[1][pos/GEN_SIZE] = 1;
		break;
		case 'g':
			if(gens[2])
				//gens[2][pos/GEN_SIZE] += amt;
				gens[2][pos/GEN_SIZE] = 1;
		break;
		case 't':
			if(gens[3])
				//gens[3][pos/GEN_SIZE] += amt;
				gens[3][pos/GEN_SIZE] = 1;
		break;
		case '-':
			if(gens[4])
				//gens[3][pos/GEN_SIZE] += amt;
				gens[4][pos/GEN_SIZE] = 1;
		break;
		case 'n':
			if(gens[5])
				//gens[3][pos/GEN_SIZE] += amt;
				gens[5][pos/GEN_SIZE] = 1;
		break;
	}
}

string Genome::GetPos(const pair<unsigned long,bool> p_pos) {
	return GetPos(p_pos.first,p_pos.second);
}
string Genome::GetPos(const unsigned long &pos, const bool upstrand) {
	int chr_num = 0;

	while(names[chr_num].second <= pos && ((unsigned int)chr_num < names.size()) )
		chr_num++;
		
	ostringstream to_return;
	
	chr_num--;
	to_return << names[chr_num].first << ":";
	if(upstrand)
		to_return << "+";
	else
		to_return << "-";
	to_return << (pos - names[chr_num].second);
	
	return to_return.str();
}

int Genome::getSetThreadID(long unsigned int pos, pthread_t myID) {
	// Already been set
	if(genome[pos/GEN_SIZE].thread_id == myID)
		return -1;
	
	// Hasn't been set so set it.
	genome[pos/GEN_SIZE].thread_id = myID;
	return 0;
}

void Genome::PrintFinal(const char* fn) {
	printf("GEN SIZE: %d\n",GEN_SIZE);
	if(!reads[0]) {	//normal sgr
		printf("*****Normal SGR*******\n");
		char* filename = new char[100];
		strcpy(filename,fn);  
		ofstream out(strcat(filename,".sgr"));
		
		for(unsigned int i=0,count=0; i<names.size()-1; i++) {
			for(; count<names[i+1].second; count+=GEN_SIZE) {
				if(genome[count/GEN_SIZE].amount> 0)
					out << names[i].first << "\t" << count - (names[i].second) 
						<< "\t" << (genome[count/GEN_SIZE].amount) << endl;
			}
		}
		out.close();
		delete[] filename;
	}

	if(reads[0]) {	// if any of the reads have been instantiated...
		// We got a new update from Evan.  Instead of printing out a file for each, we want to print
		// out one file (dubbed the '.sgrex' file for 'sgr-extra') to be a tab-deliminated file
		// for the number of a's, c's, g's, t's, and gaps at each location in the genome.
		char sgrex_fn[strlen(fn)+10];
		strcpy(sgrex_fn,fn);
		strcat(sgrex_fn,".sgrex");
		printf("File is: %s\n",sgrex_fn);
		FILE* sgrex_file = fopen(sgrex_fn,"w");
		if(!sgrex_file)
			perror("Could not open .sgrex file");
		// output the Header
		fprintf(sgrex_file,"chr\tpos\tsgr_val\tread_a\tread_c\tread_g\tread_t\tread_gap\n");
		
		if(VERBOSE > 1) fprintf(stderr,"NAMES: %ud\n",(unsigned int)names.size());

		for(unsigned int j=0,count=0; j<names.size()-1; j++) {
			if(VERBOSE > 1) fprintf(stderr,"\t[%d]: %s\n",j,names[j].first.c_str());
			for(; count<names[j+1].second; count+=GEN_SIZE) {
				if(genome[count/GEN_SIZE].amount > (float)0.0) {	//if something matched here...
									 //  chr pos  sgr
					fprintf(sgrex_file, "%s\t%ld\t%f", names[j].first.c_str(), count-names[j].second,
								genome[count/GEN_SIZE].amount);

					// Print out each of the read positions
					for(unsigned int read_count=0; read_count<NUM_CHAR_VALS; read_count++) {
						fprintf(sgrex_file,"\t%f",reads[read_count][count/GEN_SIZE]);
					}

					// Carriage return
					fprintf(sgrex_file,"\n");
				}
			}
		}
		
		// Close the file
		fclose(sgrex_file);
	}
	/*
	for(unsigned i=0; i<NUM_CHAR_VALS; i++) {
		if(reads[i]) {
			char readfile[100+6];
			strcpy(readfile,fn);

			strcat(readfile,".");
			if(i == 0)
				strcat(readfile,"a");
			else if(i == 1)
				strcat(readfile,"c");
			else if(i == 2)
				strcat(readfile,"g");
			else if(i == 3)
				strcat(readfile,"t");
			else if(i == 4)
				strcat(readfile,"-");	//For the gap

			strcat(readfile,".sgr");
			out.open(readfile);
			for(unsigned int j=0,count=0; j<names.size()-1; j++) {
				for(; count<names[j+1].second; count+=GEN_SIZE) {
					if(reads[i][count/GEN_SIZE] > 0)
						out << names[j].first << "\t" << count - (names[j].second) << "\t" << reads[i][count/GEN_SIZE] << endl;
					//if(a_spots[count/GEN_SIZE] > 0)
						//out << names[i].first << "\t" << count - (names[i].second) << "\t" << a_spots[count/GEN_SIZE] << endl;
				}

			}
			out.close();
		}

		if(gens[i]) {
			char genfile[100+10];
			strncpy(genfile,fn,100);
			strcat(genfile,".gen.");
			if(i == 0)
				strcat(genfile,"a");
			else if(i == 1)
				strcat(genfile,"c");
			else if(i == 2)
				strcat(genfile,"g");
			else if(i == 3)
				strcat(genfile,"t");
			else if(i == 4)
				strcat(genfile,"-");	//For the gap

			strcat(genfile,".sgr");
			out.open(genfile);
			for(unsigned int j=0,count=0; j<names.size()-1; j++) {
				for(; count<names[j+1].second; count+=GEN_SIZE) {
					if(gens[i][count/GEN_SIZE] > 0)
						out << names[j].first << "\t" << count - (names[j].second) << "\t" << gens[i][count/GEN_SIZE] << endl;
					//if(a_spots[count/GEN_SIZE] > 0)
						//out << names[i].first << "\t" << count - (names[i].second) << "\t" << a_spots[count/GEN_SIZE] << endl;
				}

			}
			out.close();
		}
	}
	*/
}

unsigned long Genome::size() {
	return genome_size * GEN_SIZE;
}

bool Genome::Test(ostream &os) {
	bool success = true;
	 
	unsigned char* genome = new unsigned char[1000000];
	ifstream in("test/short.fa");
	string name;
	getline(in,name);
	in.read((char*)genome,1000000);
	
	//&genome.front() = "actgctgccggtN\ncccgtcggcaacgtgNNNNNgaa\n\0";
	
	try {
		os << "**TESTING HASH_AND_STORE**\n";

		Genome gen45;
		gREAD = true;
		gREAD_FN = "/data/public/genomes/human/gnumap/gnumap_human_s2_h100k_m11";
		gen45.use("/data/public/genomes/human/gnumap/gnumap_human_s2_h100k_m11");
		gen45.LoadGenome();
		os << "Getting string at the end of chr4: " << 1877852952-200 << " " << gen45.GetPos(1877852952-200,true) << endl;
		os << gen45.GetString(1877852952-200,200) << endl;
		os << "Getting string at the end of chrX: " << 2866253375-200 << " " << gen45.GetPos(2866253375-200,true) << endl;
		os << gen45.GetString(2866253375-200,200) << endl;
		gen45.AddScore(2866253375,4.76);
		os << "Getting value at 2866253375: " << gen45.GetScore(2866253375) << endl;
		os << "                    Should be 4.76" << endl;
		gREAD = false;
		
	}
	catch(const char* erro) {
		os << "\t********Error! Did you include the proper test files?" << endl;
	}
	catch (...) {
		os << "Unknown Error!" << endl;
	}

	delete[] genome;
	return success;
}
