#include "SeqReader.h" 

SeqReader::SeqReader() :
	seq_counter(0), offset(0), this_filename("") {

	for(unsigned int i=0; i<READ_BUFFER; i++) {
		reads[i] = 0;	//clear it all
		//reads[i].pwm = 0;
	}

}

SeqReader::SeqReader(const char* fn) :
	seq_counter(0), offset(0), this_filename("") {
	
	clearReads(reads);
	
	//Separate the different files.
	char files[strlen(fn)];
	strcpy(files,fn);
	
	const char* delims = " \"\n,";
	char* token = strtok(files,delims);
	if(VERBOSE > 1)
		cout << "Found:" << endl;

	while(token != NULL) {
		if(VERBOSE > 1)
			cout << "\t" << token << endl;
		string temp_str = token;
		seq_files.push_back(temp_str);
		
		token = strtok(NULL,delims);
	}
	
}

SeqReader::~SeqReader() {
	//in.close();
	clearReads(reads);
}

void SeqReader::clearReads(Read* readArr[]) {
	if(!readArr)
		return;

	// changing our mind.  We don't want to delete everything here.
	// we want the SeqDriver to delete them after he's done with them.
	/*Read* tempR = readArr;
	while(tempR->pwm) {
		tempR = 0;
		for(unsigned int i=0; i<tempR->length; i++) {
			delete[] tempR->pwm[i];
		}
		delete[] tempR->pwm;
		delete tempR;	//don't delete this reference.  Keep the array around.
		tempR++;
	}*/
	for(unsigned int i=0; i<READ_BUFFER; i++) {
		readArr[i] = 0;
	}
	num_reads = 0;
}

void SeqReader::Init(const char* fn) {
	offset=0;	
	seq_counter=0;
	//reads.clear();
	if(reads[0])
		clearReads(reads);

	find_type();
	
	if(this_type == PRB)
		get_more_prb();
	else
		get_more_int();
	
}

void SeqReader::find_type() {
	ifstream infile(this_filename.c_str());
	string temp;
	getline(infile,temp, '\t');
	
	//The _int.txt file will have 4 tab-delimintated numbers before any spaces.
	//The _prb.txt file (created by Illumina's pipeline) will have 4 numbers--
	//  separated by spaces--before any tabs.
	string::size_type loc = temp.find(' ',0);
	if(loc != string::npos)
		this_type = PRB;
	else
		this_type = INT;
}

void SeqReader::use(string &str) {
	use(str.c_str());
}
void SeqReader::use(const char* fn) {
	this_filename = fn;
	Init(fn);
}

Read* SeqReader::GetNextSequence() {
	bool has_more=true;
	
	//if(seq_counter >= reads.size()) {
	if(seq_counter >= num_reads) {
		if(this_type == PRB)
			has_more = get_more_prb();
		else
			has_more = get_more_int();
	}
	
	if(!has_more) {
		return NULL;
	}

	return reads[seq_counter++]; 
}

//vector<vector<double> > SeqReader::PeekNextSequence() {
Read* SeqReader::PeekNextSequence() {
	//if(seq_counter > reads.size()) {
	if(seq_counter > num_reads) {
		if(this_type == PRB)
			get_more_prb();
		else
			get_more_int();
	}
	
	if(seq_counter >= num_reads || seq_counter == 0)
		return NULL;
	
	return reads[seq_counter+1];
}

bool SeqReader::get_more_prb() {

	clearReads(reads);
	//reads.clear();
	seq_counter = 0;

	ifstream in;
	in.open(this_filename.c_str());
	in.seekg(offset);
	
	//if( !in || in.eof()) {
	if(in.eof())
		return false;

	char* token;
	const char* delims = "\n\r\t ";

#ifdef DEBUG
	if(VERBOSE > 1)
		cout << "Reading more reads" << endl;
#endif
	
	//READ_BUFFER is the amount of reads we'll read into memory.
	for(unsigned int to_read = 0; to_read < READ_BUFFER; to_read++) {

#ifdef DEBUG
		if(VERBOSE > 1 && (to_read % 100 == 0 ) )	{
			cout << ".";
			cout.flush();
		}
#endif

		vector< vector<double> > one_read;

		char temp_chr[BUFFER_SIZE];
		in.getline(temp_chr,BUFFER_SIZE);
		offset += in.gcount();
		if(in.eof()) {
			if(to_read > 0)
				return true;
			return false;
		}
			

		token = strtok(temp_chr,delims);
		int first = 1;

		while(token != NULL) {
			vector<double> one_char(4);
	
			char* a_check;
			if(first) {
				a_check = token;
				first = 0;
			}
			else
				 a_check = strtok(NULL,delims);

			char* c_check = strtok(NULL,delims);
			char* g_check = strtok(NULL,delims);
			char* t_check = strtok(NULL,delims);

			//If they're all NULL, then we're done
			if(!a_check && !c_check && !g_check && !t_check)
				break;

			//this would only have one not being the right length.
			//that should throw an error--they all should be nothing.
			if( (!a_check) || (!c_check) || (!g_check) || (!t_check) )
				throw "Bad Sequence Probability File";
				
			double a = atof(a_check);
			double c = atof(c_check);
			double g = atof(g_check);
			double t = atof(t_check);
			
#ifdef DEBUG
			if(VERBOSE > 2)
				cout << "a: " << a << "\tc: "  << c << "\tg: " << g << "\tt: "  << t << endl;
#endif

			a = Q2Prb(a);
			c = Q2Prb(c);
			g = Q2Prb(g);
			t = Q2Prb(t);

			one_char[0] = a;
			one_char[1] = c;
			one_char[2] = g;
			one_char[3] = t;

			one_read.push_back(one_char);
			
		}
/*
#ifdef DEBUG
		if(VERBOSE > 3) {
			string consensus = GetConsensus(one_read); 
			cout << "consensus: " << consensus << "\t";
			bin_seq bs;
			cout << "max align: " << bs.get_align_score(one_read,consensus) << endl;
		}
#endif
*/
		
		//  if there was an adaptor sequence defined, remove it from the end
		//  of this read.
		if(adaptor)
			FixReads(one_read);
			
		//Read* temp_read = new Read;
		reads[num_reads] = new Read;
		reads[num_reads]->length = one_read.size();
		reads[num_reads]->pwm = new float*[one_read.size()];

		for(unsigned int i=0; i<one_read.size(); i++) {
			reads[num_reads]->pwm[i] = new float[4];
			for(unsigned int j=0; j<4; j++) {
				reads[num_reads]->pwm[i][j] = (float)one_read[i][j];
			}
		}
#ifdef DEBUG
		for(unsigned int q=0; q<reads[num_reads]->length; q++) {
			printf("%f %f %f %f\n",reads[num_reads]->pwm[q][0],reads[num_reads]->pwm[q][1],
								   reads[num_reads]->pwm[q][2],reads[num_reads]->pwm[q][3]);
		}
		string consensus = GetConsensus(reads[num_reads]); 
		printf("Length: %d\nconsensus: %s\n",reads[num_reads]->length,consensus.c_str());
		//cout << "consensus: " << consensus << "\t";
		printf("\n\n");
#endif
		
		num_reads++;
		//&reads[num_reads++] = temp_read;
		//reads.push_back(one_read);
	}
	
	in.close();
	
	return true;

}

inline double SeqReader::Q2Prb(double Q) {
	double answer = 1 - 1/ (pow( 10.0,(Q/10) ) + 1);
	
	return answer;
}

bool SeqReader::get_more_int() {
	clearReads(reads);
	//reads.clear();
	seq_counter = 0;
	ifstream in;
	in.open(this_filename.c_str());
	in.seekg(offset);
	
	if(in.eof())
		return false;

	char* token;
	const char* delims = "\t ";

	if(VERBOSE > 1)
		cout << "Reading more reads" << endl;
	
	//READ_BUFFER is the amount of reads we'll read into memory.
	for(unsigned int to_read = 0; to_read < READ_BUFFER; to_read++) {
#ifdef DEBUG
		if(VERBOSE > 1 && (to_read % 100 == 0 ) )	{
			cout << ".";
			cout.flush();
		}
#endif
		vector< vector<double> > one_read;

		//string temp;
		//getline(in,temp);

		char temp_chr[BUFFER_SIZE];
		in.getline(temp_chr,BUFFER_SIZE);
		offset += in.gcount();
		if(in.eof()) {
			if(to_read > 0)
				return true;
			return false;
		}

#ifdef DEBUG
		if(VERBOSE > 2)
			cout << temp_chr << endl;
#endif
			
		token = strtok(temp_chr,delims);

		//There are 4 params that don't matter to us...
		for(int i=0; i<4; i++)
			token = strtok(NULL,delims);

		int first = 1;

		while(token != NULL) {
			vector<double> one_char(4);

			char* a_check;
			if(first) {
				a_check = token;
				first = 0;
			}
			else
				 a_check = strtok(NULL,delims);

			char* c_check = strtok(NULL,delims);
			char* g_check = strtok(NULL,delims);
			char* t_check = strtok(NULL,delims);
			
			if(!a_check && !c_check && !g_check && !t_check)
				break;

			//this would only have one not being the right length.
			//that should throw an error--they all should be nothing.
			if( (!a_check) || (!c_check) || (!g_check) || (!t_check) )
				throw "Bad Sequence Intensity File";
			
			double a = atof(a_check);
			double c = atof(c_check);
			double g = atof(g_check);
			double t = atof(t_check);
			
			//The next few lines deal with a negative sequence probability,
			//so that the probability of a base is never > 1 
			//
			double shift = 0;
			if(a < shift)
				shift = a;
			if(c < shift)
				shift = c;
			if(g < shift)
				shift = g;
			if(t < shift)
				shift = t;
				
			if(shift < 0) {
				shift = shift * -1;
				a += shift;
				c += shift;
				g += shift;
				t += shift;
			}
			
			double total = a + c + g + t;
			
			if(total != 0) {
				a /= total;
				c /= total;
				g /= total;
				t /= total;
			}
#ifdef DEBUG
			if(VERBOSE > 2)
				cout << "a: " << a << "\tc: "  << c << "\tg: " << g << "\tt: "  << t << endl;
#endif
			one_char[0] = a;
			one_char[1] = c;
			one_char[2] = g;
			one_char[3] = t;

			one_read.push_back(one_char);
			//one_read[i] = one_char;
			//delete[] begin;
		}
/*
#ifdef DEBUG
		if(VERBOSE > 3) {
			string consensus = GetConsensus(one_read); 
			cout << "consensus: " << consensus << "\t";
			bin_seq bs;
			cout << "max align: " << bs.get_align_score(one_read,consensus) << endl;
			
		}
#endif
*/

		//  if there was an adaptor sequence defined, remove it from the end
		//  of this read.
		if(adaptor)
			FixReads(one_read);

		//Read* temp_read = new Read;
		reads[num_reads] = new Read;
		reads[num_reads]->length = one_read.size();
		reads[num_reads]->pwm = new float*[one_read.size()];

		for(unsigned int i=0; i<one_read.size(); i++) {
			reads[num_reads]->pwm[i] = new float[4];
			for(unsigned int j=0; j<4; j++) {
				reads[num_reads]->pwm[i][j] = (float)one_read[i][j];
			}
		}
		num_reads++;
		//&reads[num_reads++] = temp_read;
		//reads.push_back(one_read);
	}
	
	in.close();
	
	return true;
}

float SeqReader::Compare(string consense, string adaptor, int pos) { 
	unsigned int num_same=0, num_diff=0;

	unsigned int j,i;
	for(i=pos,j=0; j<adaptor.length() && i<consense.length(); i++,j++) {
		if(consense[i] == adaptor[j])
			num_same++;
		else
			num_diff++;
	}

	float perc_identity = (float)num_same / j;
	return perc_identity;
}

void SeqReader::FixReads(vector<vector<double> > &m_read) { 
//void SeqReader::FixReads(Read &m_read) { 
	string s_adaptor = adaptor;
	string consense = GetConsensus(m_read);
	unsigned int c_len = consense.length();
	//unsigned int a_len = s_adaptor.length();

	//if(c_len-a_length < HASH_SIZE)
	//	return;	//don't adapt it if it'll be shorter than the hash size

	//Chop Adaptor length here.
	// we only are worried about the last *i* characters of the sequence
	//int i;
	//for(i=a_len-1; i>0; i--) {
	//	if(consense.substr(c_len-i) == s_adaptor.substr(0,i))
	//		break;
	//}
	unsigned int i;
	float diff=0;
	for(i=0; i<c_len-MIN_CHOPPED_BASES; i++) {
		diff = Compare(consense,s_adaptor,i);
		if(diff >= MIN_ADAPTOR_DIFF)
			break;
	}

#ifdef DEBUG
	if(VERBOSE > 1)
		printf("Chopping read %s at pos %d (chopped %d bases) with diff %f\n",consense.c_str(),i,c_len-i,diff);
#endif
	// if there was a portion of the adaptor that matched the read,
	// pop it off the read.
	if( (c_len-i) > MIN_CHOPPED_BASES) {
		//printf("Fixing read by %d\n",c_len-i); 
		for(unsigned int j=c_len; j>i; j--) {
			m_read.pop_back();
			//delete[] m_read[i];
			//m_read->length--;
		}
	}
	//don't return anything.  Modify the *read* reference.

#ifdef DEBUG
	if(VERBOSE > 1) {
		printf("              %*s\n",i+(int)strlen(adaptor),adaptor);
		printf("              %s\n",GetConsensus(m_read).c_str());
	}
#endif

}

string SeqReader::GetConsensus(Read* read) {
	vector<vector<double> > pwm;
	for(unsigned int i=0; i<read->length; i++) {
		vector<double> one_chr;
		for(unsigned int j=0; j<4; j++) {
			one_chr.push_back(read->pwm[i][j]);
		}
		pwm.push_back(one_chr);
	}
	
	return GetConsensus(pwm);
}

string SeqReader::GetConsensus(vector<vector<double> > &read) {
//string SeqReader::GetConsensus(Read &pwm) {
	string seq = "";

	for(unsigned int i=0; i<read.size(); i++) {
		seq += max_char(read[i]);
	//for(unsigned int i=0; i<read.length; i++) {
	//	seq += max_char(read.pwm[i]);
	}

	return seq;
}

/*
 * Because this is used to produce the consensus sequence, we need ambiguity characters...
 */
inline char SeqReader::max_char(vector<double> &chr) {
//inline char SeqReader::max_char(float chr[4]) {
	if((chr[0] == chr[1]) && (chr[0] == chr[2]) && (chr[0] == chr[3]))
		return 'n';
	if(chr[0] >= chr[1]) {
		if(chr[0] >= chr[2]) {
			if(chr[0] >= chr[3])
				return 'a';
			else
				return 't';
			}
		else {
			if(chr[2] >= chr[3])
				return 'g';
			else
				return 't';
		}
	}
	else {
		if(chr[1] >= chr[2]) {
			if(chr[1] >= chr[3])
				return 'c';
			else
				return 't';
		}
		else {
			if(chr[2] >= chr[3])
				return 'g';
			else return 't';
		}
	}

	return 'n';
}

unsigned int SeqReader::GetNumSeqs() {
	ifstream inf(this_filename.c_str());
	inf.clear();
	
	unsigned int line_count=0;
	
	char temp[BUFFER_SIZE];
	
	while(inf) {
		inf.getline(temp, BUFFER_SIZE);
		if(inf.eof())
			break;
		line_count++;
	}
		
	return line_count;
}

unsigned int SeqReader::GetNumSeqFiles() {
	return seq_files.size();
}

vector<string>& SeqReader::GetSeqFileVector() {
	return seq_files;
}
		
bool SeqReader::Test(ostream &os) {
	bool success = true;

	try {
		SeqReader sr;
		sr.use("test/test_int.txt");

		unsigned int i=0;
		//while(sr.GetNextSequence().size() > 0)
		while(sr.GetNextSequence())
			i++;
		cout << "I: " << i << endl;
		TEST(i == sr.GetNumSeqs());
		/*
		TEST(sr.max_char(1.0, 2.2, 3.3, 4.4) == 't');
		TEST(sr.max_char(1.0, .9, .8, .7) == 'a');
		TEST(sr.max_char(1.0, 2.0, 1.5, 1.6) == 'c');
		TEST(sr.max_char(1.1, 1.2, 1.3, 1.4) == 't');
		TEST(sr.max_char(1.1, 1.2, 1.3, 1.25) == 'g');
		TEST(sr.max_char(0, 0, 0, 0) == 'a')
		TEST(sr.max_char(0, 0, 0, 4) == 't')
		TEST(sr.max_char(1, 5, 0, 4) == 'c')
		TEST(sr.max_char(5, 1, 4, 4) == 'a')
		TEST(sr.max_char(1, 4, 4, 4) == 'c')
		TEST(sr.max_char(0, 4, 0, 3) == 'c')
		*/

		os << "*********NEEDS MORE***********\n";
	}
	catch(const char* error) {
		os << "\t********Error! Did you include the proper test files?" << endl;
	}
	return success;
}

