#include "bin_seq.h"
	
#define NEG_INF -1000;

bin_seq::bin_seq() {}

/*
 * The following functions are used for hashing the genome.  Thus the name "binary sequence"
 */
unsigned int bin_seq::str2int(string &a) {
	unsigned int temp = 0;
	for(unsigned int i=0; i<16 && i<a.size(); i++) {
		temp <<= 2;
		//temp += base2int(a[i]);
		temp |= bs_CONVERSION[(int)a[i]];
	}
	
	return temp;
}


string bin_seq::hash2str(const unsigned int h, int length) {
	char to_return[length];
	unsigned int temp = h;

	for(int i=1; i<=length; i++) {
		//to_return[length-i] = int2base((temp % 4));
		to_return[length-i] = INT2BASE[(temp % 4)];
		temp >>= 2;
	}
	
	to_return[length] = '\0';
	
	return to_return;
}

pair<bool,unsigned long> bin_seq::get_hash(const char* str) {
	string str_str = str;
	
	return get_hash(str_str);
}

pair<bool,unsigned long> bin_seq::get_hash(string str) {
	unsigned long temp = 0;

	//ensure that the sequence doesn't overflow the possible size of an unsinged int,
	//as well as not overflowing the MER_SIZE.
	for(unsigned int i=0; i<(sizeof(unsigned long) * 4) && i<str.length(); i++) {
		
		//int spot = base2int(str[i]); 
		int spot = bs_CONVERSION[(int)str[i]]; 
		if(spot > 3)	//if it's not a valid, hashable character
			return pair<bool,unsigned long>(false,0);
			
		temp <<= 2;
		temp |= spot;
	}
		
	//temp %= HASH_SIZE;
	return pair<bool,unsigned long>(true,temp);
}

pair<int,unsigned long> bin_seq::get_hash(unsigned int &pos, const unsigned char* genome) {
	unsigned int temp = 0;

	int found = 0;
	// Ensure that the sequence doesn't overflow the possible size of a unsigned int,
	// as well as not overflowing the MER_SIZE.  (We won't allow MER_SIZE to
	// overflow the size of an unsigned long--64 bits.  This means we can allow up to 32bp MER_SIZE)
	// We want to make sure we find exactly MER_SIZE bases (found), but the int i helps us determine
	// where on the genome we look for them--while ignoring newline characters.
	//for(int i=0,found=0; found<MER_SIZE && found<64 && pos+i<BUFFER_SIZE; found++, i++) {
	for(int i=0; found<MER_SIZE; i++) {
		//unsigned int spot = base2int(genome[pos+i]);
		unsigned int spot = bs_CONVERSION[(int)genome[pos+i]];
		
		if(spot < 4) {
			temp <<= 2;
			temp |= spot;
		
			found++;
		}
		else if(spot == 4) {	//if there's an newline reached, this isn't an error: just ignore.
			if(i == 0) {	//if there's a newline reached on the first of a sequence...
				pos++;
				i=-1;
			}
		}
		else if(spot == 5) {
			//pos = BUFFER_SIZE;	// 5 comes from the flag for the end of the stream.
			return pair<int,unsigned int>(-1,0);
		}
		else if(spot == 7) {
			pos = pos+i;
			//return a flag for a new genome, with a number representing the number of bases skipped.
			return pair<int,unsigned int>(-2,found); 
		}
		else { // if(spot == 6) 
			//cout << "Unknown character: <" << genome[pos+i] << ">" << endl;
			return pair<int,unsigned int>(0,0);
		}
	}
		
	return pair<bool,unsigned int>(1,temp);
}

inline unsigned int bin_seq::base2int(char c) {
	switch(c) {
		case 'A' : return 0;
		case 'C' : return 1;
		case 'G' : return 2;
		case 'T' : return 3;
		case 'a' : return 0;
		case 'c' : return 1;
		case 'g' : return 2;
		case 't' : return 3; 
		case '\n' : return 4;
		case '\r' : return 4;
		case '\0' : return 5;	//5 is the flag for the end of the genome stream.
		case 'N' : return 6;
		case 'n' : return 6;
		case '>' : return 7;	//7 means there's a new genome.
	}
	return 6;
} 
inline char bin_seq::int2base(unsigned int i) {
	switch(i) {
		case 0 : return 'a';
		case 1 : return 'c';
		case 2 : return 'g';
		case 3 : return 't';
	}
	return 'N';
} 



/*
 * The next few functions are for producing a Needleman-Wunsch alignment
 * for two given sequences.
 */
void bin_seq::print_align(string str1, string align, string str2) {
	int s1_size = str1.size();
	for(int i=1; i<=s1_size; i++) 
		cout << str1[s1_size-i]; 
	cout << endl;
	
	int a_size = align.size(); 
	for(int i=1; i<=a_size; i++)
		cout << align[a_size-i];
	cout << endl;
	
	int s2_size = str2.size();
	for(int i=1; i<=s2_size; i++)
		cout << str2[s2_size-i];
	cout << endl;
}

//will return the alignment string
//string bin_seq::get_align_score_w_traceback(const vector< vector<double> > &pwm, const string &consense, const string &gen) {
string bin_seq::get_align_score_w_traceback(const Read &read, const string &consense, const string &gen) {
	/* 
	cout << "aligning " << gen << " with " << pwm[0][0] << "\t" << pwm[0][1] << endl;
	cout << "Scores: " << get_val(pwm[0],'a') << ' '
		 			   << get_val(pwm[0],'c') << ' '
		 			   << get_val(pwm[0],'g') << ' '		 			   
		 			   << get_val(pwm[0],'t') << endl; 
	//*/

	//int SEQ_LENGTH = pwm.size();
	int SEQ_LENGTH = read.length;

	float nm[SEQ_LENGTH+1][gen.size()+1];
	
	//Initialize the entire array to zeros.
	for(int i=0; i<=SEQ_LENGTH; i++)
		//for(int j=0; j<=gen.size(); j++) {
		for(int j=i-MAX_GAP-1; j<=i+MAX_GAP+1; j++) {
			if(j < 0 || j > (int)gen.size() )
				continue;
			nm[i][j] = NEG_INF;
		}
	
	//Initialize the first row to have the gap penalty.
	for(int i=0; i<=MAX_GAP+2; i++) {
		nm[i][0] = GAP * i;
	}
	//Initialize the first column to have the gap penalty.
	for(int j=0; j<=MAX_GAP+2; j++) {
		nm[0][j] = GAP * j;
	}

	//Set up the NW matrix	
	for(int i=1; i<=SEQ_LENGTH; i++) {
		//for(unsigned int j=1; j<=gen.size(); j++) {
		for(int j=i-MAX_GAP; j<=i+MAX_GAP; j++) {
			if(j <= 0)
				continue;
			if (j>(int)gen.size())
				break;
			
			float m_mm1 = nm[i-1][j-1];
			float m_mm2 = get_val(read.pwm[i-1],gen[j-1]);
			float m_mm = m_mm1 + m_mm2;
			float gap1 = nm[i][j-1] + GAP;
			float gap2 = nm[i-1][j] + GAP;
			
			float max_score = max_flt(m_mm,gap1,gap2);
			
			nm[i][j] = max_score;

		}	
	}
	
#ifdef DEBUG
///*	//Uncomment this to print the NW matrix
 	for(int i=0; i<=SEQ_LENGTH; i++) {
		for(int j=0; j<=gen.size(); j++) {
			cout << nm[i][j] << '\t';
			
		}
		cout << endl;
	}
	cout << endl;
// */
#endif
	
	//Trace back though the NW matrix to find the best alignment.
	string aligned = "";
	float traceback=0;
	int gaps=0;

	int i = SEQ_LENGTH;
	int j = gen.size();
		
	while( (i != 0) && (j != 0) ) {
		float match_val = get_val(read.pwm[i-1],gen[j-1]);
		float gap_val1 = GAP;
		float gap_val2 = GAP;
		//double gap_val1 = get_val(pwm[i-1],gen[j-1]);
		//double gap_val2 = get_val(pwm[i-2],gen[j-1]);
		
		//double change = nm[i-1][j-1] + match_val;
		//double change_g1 = nm[i][j-1] + gap_val1;
		//double change_g2 = nm[i-1][j] + gap_val2;
		//There's a chance for a round-off error in this shebang.
		//if( (change < 0.001) && (change > -0.001) ) {
		//if( ( (float)(nm[i-1][j-1] + match_val) == (float)nm[i][j]) ) {
		if( ( (nm[i-1][j-1] + match_val) == nm[i][j]) ) {

			aligned += consense[j-1];
			traceback += match_val;
			//if(consense[j-1] == 'c') {
			//	if(gen[i-1] == 'c')
			//		c2c++;
			//	else if(gen[i-1] == 't')
			//		c2t++;
			//}
			i -= 1;
			j -= 1;
		}
		//else if( ( (float)(nm[i][j-1] + gap_val1) == (float)nm[i][j]) ) {
		else if( ( (nm[i][j-1] + gap_val1) == nm[i][j]) ) {
			traceback += gap_val1;
			aligned += consense[j];
			j -= 1;
			gaps++;
		}
		//else if( ( (float)(nm[i-1][j] + gap_val2) == (float)nm[i][j]) ) {
		else if( ( (nm[i-1][j] + gap_val2) == nm[i][j]) ) {
			traceback += gap_val2;
			aligned += "-";
			i -= 1;
			gaps++;
		}
		else {
			cout << "nm + match: " << nm[i-1][j-1] + match_val << "\tnm: " << nm[i][j] << endl;
			cout << "nm + match_g1: " << nm[i][j-1] + match_val << "\tnm: " << nm[i][j] << endl;
			cout << "nm + match_g2: " << nm[i-1][j] + match_val << "\tnm: " << nm[i][j] << endl;
			cout << "Error at " << i << " , " << j << " ." << endl;
			return "";
		}

		if(gaps > MAX_MISMATCH) {
			//cout << "too many gaps." << endl;
			return "";
		}
		
	}

#ifdef DEBUG
	if(VERBOSE > 3) {
		cout << "final answer: " << traceback << endl;
		cout << "gaps: " << gaps << endl;
	}
#endif

	//Get the reverse of this sequence
	string temp = aligned; 
	for(unsigned int i=0; i<aligned.size(); i++) {
		temp[i] = aligned[aligned.size()-1-i];
	}
	aligned = temp;
	
	return aligned;
}

/*
 * Puts all the functions together and returns the value incorporated here.
 */
float bin_seq::get_align_score(Read &read, string &gen,
		unsigned int begin, unsigned int end) {
		
	float value = 0;
	
	value += get_align_score_begin(read,gen,begin);
	value += get_align_score_mid(read,gen,begin,end);
	value += get_align_score_end(read,gen,end);
	
	return value;
		
}

/*
 * This is meant to align the first piece of the read.  Aligning from the (exact) matching
 * portion which was produced from the hash piece to the beginning of the read.
 *
 *  *
 *	  *		<- we want to match this first piece.
 *		*
 *        *
 *          X	<- end.  We know it must be an exact match
 *			  X
 *
 */
float bin_seq::get_align_score_begin(const Read &read, const string &gen, unsigned int end) {
	if(end == 0)	//trivial case
		return 0;
		
	float nm[end+1][end+1];
	
	//Initialize the middle of the array to zeros.
	for(unsigned int i=0; i<=end; i++) {
		for(int j=(signed int)i-MAX_GAP-1; j<=(signed int)i+MAX_GAP+1 && j<=(signed int)end; j++) {
			if(j < 0)
				continue;
			nm[i][j] = NEG_INF;
		}
	}
	
	//Initialize the last row to have the gap penalty.
	//for(int i=MAX_GAP+2; i<=(signed int)end; i++) {
	for(int i=0; i<=(signed int)end; i++) {
		nm[i][end] = GAP * (end-i);
	}
	//Initialize the last column to have the gap penalty.
	//for(int j=MAX_GAP+2; j<=(signed int)end; j++) {
	for(int j=0; j<=(signed int)end; j++) {
		nm[end][j] = GAP * (end-j);
	}
	
	
	//Set up the NW matrix	
	for(int i=end-1; i>=0; i--) {	//needs to be signed when it goes to 0...
		for(int j=i+MAX_GAP; j>=i-MAX_GAP; j--) {	//j needs to be signed
			if (j>=(signed int)end) 
				continue;
			if (j<0)
				break;


#ifdef DEBUG
/*	
	for(int q=0; q<=end; q++) {
		for(int p=0; p<=end; p++) {
			cout << nm[q][p] << '\t';
			
		}
		cout << endl;
	}
	cout << endl;
*/
#endif

			float m_mm1 = nm[i+1][j+1];
			float m_mm2 = get_val(read.pwm[i],gen[j]);
			float m_mm = m_mm1 + m_mm2;
			float gap2 = nm[i+1][j] + GAP;
			float gap1 = nm[i][j+1] + GAP;
			//double gap1 = nm[i][j-1] + get_val(pwm[i],gen[j]);
			//double gap2 = nm[i-1][j] + get_val(pwm[i-1],gen[j-1]);
			
			float max_score = max_flt(m_mm,gap1,gap2);
			
			nm[i][j] = max_score;
		}	
	}

#ifdef DEBUG
/*
	cout << "\n\nend: " << end << endl;
	for(int i=0; i<=end; i++) {
		for(int j=0; j<=end; j++) {
			cout << nm[i][j] << '\t';
			
		}
		cout << endl;
	}
	cout << endl;
*/
#endif	

	float align_score = nm[0][0];
	return align_score;
}

/*
 * We know there's an exact match, so we just compute the alignment score with itself
 * from start to end.
 * start is the first position of the exact match.  end is the last position of the 
 * exact match (not the first position of a possible match).
 */
float bin_seq::get_align_score_mid(const Read &read, const string &consense, 
			unsigned int start, unsigned int end) {
	float score = 0;
	for(unsigned int i=start; i<=end; i++) {
		score += (read.pwm[i][0] * ALIGN_SCORES[(unsigned int)consense[i]][0])
		   + (read.pwm[i][1] * ALIGN_SCORES[(unsigned int)consense[i]][1])
		   + (read.pwm[i][2] * ALIGN_SCORES[(unsigned int)consense[i]][2])
		   + (read.pwm[i][3] * ALIGN_SCORES[(unsigned int)consense[i]][3]);
	}
	
	return score;
}

/*
 * This is meant to align the last piece of the read.  Aligning from the (exact) matching
 * portion which was produced from the hash piece to the end of the read.
 *
 *  X
 *	  X		<- start.  We know it must be an exact match
 *		*
 *        *
 *          *	<- we want to match this last piece.
 *			  *
 *
 */
float bin_seq::get_align_score_end(const Read &read, const string &gen, unsigned int start) {
	if(start == read.length-1)	//trivial case
		return 0;
		
	unsigned int length = read.length - start;	//start+length == read.length
	
	float nm[length][length];
	
	//Initialize the middle of the array to zeros.
	for(unsigned int i=0; i<length; i++) {
		for(int j=i-MAX_GAP-1; j<=(signed int)i+MAX_GAP+1 && j < (signed int)length; j++) {	//j needs to be signed
			if(j < 0)
				continue;
			nm[i][j] = NEG_INF;
		}
	}
	
	//Initialize the first row to have the gap penalty.
	for(unsigned int i=0; i<=(unsigned int)MAX_GAP+1 && i<length; i++) {
		nm[i][0] = GAP * i;
	}
	//Initialize the first column to have the gap penalty.
	for(unsigned int j=0; j<=(unsigned int)MAX_GAP+1 && j<length; j++) {
		nm[0][j] = GAP * j;
	}
	//Set up the NW matrix	
	for(unsigned int i=1; i<length; i++) {
		for(int j=i-MAX_GAP; j<=(signed int)i+MAX_GAP && j<(signed int)length; j++) {
			if(j <= 0)
				continue;
			if (j>=(signed int)length)
				break;
			
			float m_mm1 = nm[i-1][j-1];
			float m_mm2 = get_val(read.pwm[i+start],gen[j+start]);
			float m_mm = m_mm1 + m_mm2;
			float gap1 = nm[i][j-1] + GAP;
			float gap2 = nm[i-1][j] + GAP;
			//double gap1 = nm[i][j-1] + get_val(pwm[i],gen[j]);
			//double gap2 = nm[i-1][j] + get_val(pwm[i-1],gen[j-1]);
			
			float max_score = max_flt(m_mm,gap1,gap2);
			
			nm[i][j] = max_score;

		}	
	}
	
	float align_score = nm[length-1][length-1];
	return align_score;
}




float bin_seq::get_align_score(Read &read, string &gen) {
//double bin_seq::get_align_score(vector< vector<double> > &pwm, string &gen) {
	/* 
	cout << "aligning " << gen << " with " << pwm[0][0] << "\t" << pwm[0][1] << endl;
	cout << "Scores: " << get_val(pwm[0],'a') << ' '
		 			   << get_val(pwm[0],'c') << ' '
		 			   << get_val(pwm[0],'g') << ' '		 			   
		 			   << get_val(pwm[0],'t') << endl; 
	//*/

	//int SEQ_LENGTH = pwm.size();
	int SEQ_LENGTH = read.length;

	float nm[SEQ_LENGTH+1][gen.size()+1];
	
	//Initialize the middle of the array to zeros.
	for(int i=0; i<=SEQ_LENGTH; i++) {
		for(int j=i-MAX_GAP-1; j<=i+MAX_GAP+1; j++) {
			if(j < 0 || j > (int)gen.size() )
				continue;
			nm[i][j] = NEG_INF;
		}
	}
	
	//Initialize the first row to have the gap penalty.
	for(int i=0; i<=MAX_GAP+2; i++) {
		nm[i][0] = GAP * i;
	}
	//Initialize the first column to have the gap penalty.
	for(int j=0; j<=MAX_GAP+2; j++) {
		nm[0][j] = GAP * j;
	}

	//Set up the NW matrix	
	for(int i=1; i<=SEQ_LENGTH; i++) {
		for(int j=i-MAX_GAP; j<=i+MAX_GAP; j++) {
			if(j <= 0)
				continue;
			if (j>(int)gen.size())
				break;
			
			float m_mm1 = nm[i-1][j-1];
			float m_mm2 = get_val(read.pwm[i-1],gen[j-1]);
			float m_mm = m_mm1 + m_mm2;
			float gap1 = nm[i][j-1] + GAP;
			float gap2 = nm[i-1][j] + GAP;
			//double gap1 = nm[i][j-1] + get_val(pwm[i],gen[j]);
			//double gap2 = nm[i-1][j] + get_val(pwm[i-1],gen[j-1]);
			
			float max_score = max_flt(m_mm,gap1,gap2);
			
			nm[i][j] = max_score;

		}	
	}
	
#ifdef DEBUG
	//Uncomment this to print the NW matrix
/* 	for(int i=0; i<=SEQ_LENGTH; i++) {
		for(int j=0; j<=gen.size(); j++) {
			cout << nm[i][j] << '\t';
			
		}
		cout << endl;
	}
	cout << endl;
*/	
	//if(VERBOSE > 3) {
	//	cout << "final answer: " << traceback << endl;
	//}
#endif

	float align_score = nm[SEQ_LENGTH][gen.size()];
	
	return align_score;
}


//inline double bin_seq::get_val(const vector<double> &a, const char b) {
inline float bin_seq::get_val(const float* a, const char b) {
	//if(VERBOSE > 4)
		//cout << "comparing " << b << endl;
	float score = 0;
	//score += (a[0] * ALIGN_SCORES[(unsigned int)b][0])
	//	   + (a[1] * ALIGN_SCORES[(unsigned int)b][1])
	//	   + (a[2] * ALIGN_SCORES[(unsigned int)b][2])
	//	   + (a[3] * ALIGN_SCORES[(unsigned int)b][3]);
	//
	//return score;

	float* scoreptr = ALIGN_SCORES[(unsigned int)b];
	score = (a[0] * scoreptr[0]
		   + a[1] * scoreptr[1]
		   + a[2] * scoreptr[2]
		   + a[3] * scoreptr[3]);
	return score;
/*
	int gen_char;
	switch(b) {
		case 'a': 
			gen_char = 0;
			break;
		case 'c':
			gen_char = 1;
			break;
		case 'g':
			gen_char = 2;
			break;
		case 't':
			gen_char = 3;
			break;
		case 'n':
			gen_char = 4;
			break;
		default:
			printf("Using default here.  Try another method.\n");
			gen_char = 4;
			break;
	}

	float score = 0;
	score += (a[0] * ALIGN_SCORES[gen_char][0])
		   + (a[1] * ALIGN_SCORES[gen_char][1])
		   + (a[2] * ALIGN_SCORES[gen_char][2])
		   + (a[3] * ALIGN_SCORES[gen_char][3]);

	return score;
*/
} 

inline float bin_seq::get_val(char a, char b) {
#ifdef DEBUG
	if(VERBOSE > 4)
		cout << "comparing " << a << " and " << b << endl;
#endif

	return ALIGN_SCORES[(unsigned int)a][(unsigned int)b];
	/*
	int first=-1,second=-1;	

	if(tolower(a) == 'a')
		first = 0;
	else if(tolower(a) == 'c')
		first = 1;
	else if(tolower(a) == 't')
		first = 2;
	else if(tolower(a) == 'g')
		first = 3;
	else
		first = 4;

	if(tolower(b) == 'a')
		second = 0;
	else if(tolower(b) == 'c')
		second = 1;
	else if(tolower(b) == 't')
		second = 2;
	else if(tolower(b) == 'g')
		second = 3;
	else 
		second = 4;

	return ALIGN_SCORES[first][second];
	*/
}

inline float bin_seq::max_flt(float a, float b, float c) {
	//cout << a << ' '<< b << ' ' << c << endl;
	if(a >= b) {
		if(a >= c)
			return a;
		else return c;	//implies c>a, which means c>b.
	}
	else {		//we know know b>a.
		if(b >= c)
			return b;
		else
			return c;
	}
}

inline int bin_seq::max(int a, int b, int c) {
	//cout << a << ' '<< b << ' ' << c << endl;
	if(a >= b) {
		if(a >= c)
			return a;
		else return c;
	}
	else {
		if(b >= c)
			return b;
		else
			return c;
	}
}

bool bin_seq::Test(ostream &os) {
	bool success = true;
	
	unsigned char* genome = new unsigned char[100000];

	ifstream in;	
	in.open("test/short.fa");
	in.seekg(7);
	in.read((char*)genome, 100000);
	
	string dna = "aaaagaaag";
	string dna2 = "aaaaaaaaa";
	string dnat = "ttttttttt";
	bin_seq bs;
	unsigned int bs2_t_good = 6285;
	unsigned int bs2_t_bad = 33;
	unsigned int good = 7057;
	bin_seq bs2;
	
	//string loads = "ccccccccc";
	//cout << "highest " <<  bs.str2int(loads) << endl;
	
	TEST(!bs.get_hash(bs2_t_bad,genome).first);
	TEST(bs.get_hash(bs2_t_good,genome).first);
	
	TEST(bs.get_hash(good,genome).second == 65712);
	TEST(bs.str2int(dna) == 514);
	//TEST(bs2.get_hash(dna) == bs.get_hash(bs2_t_good,genome));
	cout << "get hash: " << bs2.get_hash(bs2_t_good,genome).second << endl;
	cout << "dna: " << bs2.get_hash(dna).second << endl;
	TEST(bs.hash2str(514,9) == dna);
	TEST(bs.hash2str(0,9) == dna2);
	TEST(bs.hash2str(262143,9) == dnat);
	TEST(bs.get_hash(dnat).second == 262143);
	TEST(bs.get_hash("caaaagtaa").second == 65712);
	TEST(bs.get_hash("atcgccaga").second == 55624);
	
	//Used for testing the N-W alignement
	TEST(bs.max(1,2,3) == 3);
	TEST(bs.max(1,3,2) == 3);
	TEST(bs.max(3,2,1) == 3);
	TEST(bs.max(3,2,3) == 3);
	
	delete[] genome;
	in.close();
	return success;
}
